﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;

namespace Histogram_Processing_Color_Images
{
    public partial class main : Form
    {
        public main()
        {
            InitializeComponent();
            ofd = new OpenFileDialog();
            sfd = new SaveFileDialog();
            ofd.Filter = filter;
            sfd.Filter = filter;
        }

        OpenFileDialog ofd;
        SaveFileDialog sfd;
        string image_path;
        string filter = "Image | *.jpg;*.png;*.gif";
        Bitmap image, processed, display;
        public static int width, height;

        private void processToolStripMenuItem_Click(object sender, EventArgs e)
        {
            double[][] hsi = processed.RGB2HSI();
            processed = hsi.EqualizeHistogram();
            display = processed.ScaleImage(image_frame.Width, image_frame.Height);
            image_frame.Image = display;
            
        }

        private void saveToolStripMenuItem_Click(object sender, EventArgs e)
        {
            sfd.FileName = image_path;
            if (sfd.ShowDialog() == DialogResult.OK)
            {
                processed.Save(sfd.FileName);
            }
        }

        private void loadToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (ofd.ShowDialog() == DialogResult.OK)
            {
                image_path = ofd.FileName;
                image = (Bitmap)Bitmap.FromFile(image_path);
                display = image.ScaleImage(image_frame.Width, image_frame.Height);
                image_frame.Image = display;
                processed = image;
            }
        }
    }

    public static class Ext
    {
        public static Bitmap ScaleImage(this Bitmap image, int width, int height)
        {
            float xratio = (float)width / image.Width;
            float yratio = (float)height / image.Height;
            float ratio = Math.Min(xratio, yratio);

            int w = (int)(image.Width * ratio);
            int h = (int)(image.Height * ratio);

            Bitmap result = new Bitmap(image, new Size(w, h));
            return result;
        }

        public static double[][] RGB2HSI(this Bitmap image)
        {
            int w = image.Width;
            int h = image.Height;

            main.width = w;
            main.height = h;

            BitmapData image_data = image.LockBits(
                new Rectangle(0, 0, w, h),
                ImageLockMode.ReadOnly,
                PixelFormat.Format24bppRgb);
            int bytes = image_data.Stride * image_data.Height;
            byte[] buffer = new byte[bytes];
            Marshal.Copy(image_data.Scan0, buffer, 0, bytes);
            image.UnlockBits(image_data);
            double[][] result = new double[w * h][];

            for (int x = 0; x < w; x++)
            {
                for (int y = 0; y < h; y++)
                {
                    int pos = x * 3 + y * image_data.Stride;

                    int r_pos = x + y * w;
                    result[r_pos] = new double[3];
                    double b = (double)buffer[pos];
                    double g = (double)buffer[pos + 1];
                    double r = (double)buffer[pos + 2];
                    double[] pixel = { r, g, b };

                    double num = 0.5 * (2 * r - g - b);
                    double den = Math.Sqrt(Math.Pow(r - g, 2) + (r - b) * (g - b));

                    double hue = Math.Acos(num / (den));

                    if (b > g)
                    {
                        hue = 2 * Math.PI - hue;
                    }

                    num = pixel.Min();
                    den = r + g + b;
                    if (den == 0)
                    {
                        den = 0.00000001;
                    }

                    double saturation = 1d - 3d * num / den;

                    if (saturation == 0)
                    {
                        hue = 0;
                    }

                    double intensity = (r + g + b) / 3;

                    result[r_pos][0] = hue;
                    result[r_pos][1] = saturation;
                    result[r_pos][2] = intensity;
                }
            }

            //transpose

            double[][] result_t = new double[3][];

            for (int i = 0; i < 3; i++)
            {
                result_t[i] = new double[w * h];

                for (int j = 0; j < w * h; j++)
                {
                    result_t[i][j] = result[j][i];
                }
            }

            return result_t;
        }

        public static Bitmap HSI2RGB(this double[][] hsi_map)
        {
            int w = main.width;
            int h = main.height;

            Bitmap image = new Bitmap(w, h);
            BitmapData image_data = image.LockBits(
                new Rectangle(0, 0, w, h),
                ImageLockMode.WriteOnly,
                PixelFormat.Format24bppRgb);
            int bytes = image_data.Stride * image_data.Height;
            byte[] result = new byte[bytes];

            for (int x = 0; x < w; x++)
            {
                for (int y = 0; y < h; y++)
                {
                    int pos = x * 3 + y * image_data.Stride;
                    int hsi_pos = x + y * w;
                    double H = hsi_map[0][hsi_pos];
                    double S = hsi_map[1][hsi_pos];
                    double I = hsi_map[2][hsi_pos];

                    double red = 0;
                    double green = 0;
                    double blue = 0;

                    if (H >= 0 && H < 2 * Math.PI / 3)
                    {
                        blue = (I * (1 - S));
                        red = (I * (1 + S * Math.Cos(H / Math.Cos(Math.PI / 3 - H))));
                        green = (3 * I - (red + blue));
                    }

                    else if (H >= 2 * Math.PI / 3 && H < 4 * Math.PI / 3)
                    {
                        red = (I * (1 - S));
                        green = (I * (1 + S * Math.Cos(H - 2 * Math.PI / 3) / Math.Cos(Math.PI - H)));
                        blue = (3 * I - (red + green));
                    }

                    else if (H >= 4 * Math.PI / 3 && H < 2 * Math.PI)
                    {
                        green = (I * (1 - S));
                        blue = (I * (1 + S * Math.Cos(H - 4 * Math.PI / 3) / Math.Cos(5 * Math.PI / 3 - H)));
                        red = (3 * I - (green + blue));
                    }

                    if (blue > 255)
                    {
                        blue = 255;
                    }

                    if (green > 255)
                    {
                        green = 255;
                    }

                    if (red > 255)
                    {
                        red = 255;
                    }

                    result[pos] = (byte)blue;
                    result[pos + 1] = (byte)green;
                    result[pos + 2] = (byte)red;
                }
            }

            Marshal.Copy(result, 0, image_data.Scan0, bytes);
            image.UnlockBits(image_data);

            return image;
        }

        public static Bitmap EqualizeHistogram(this double[][] hsi)
        {
            double[] intensities = hsi[2];

            //normalize
            double max = 0d;
            for (int i = 0; i < intensities.Length; i++)
            {
                max = Math.Max(max, intensities[i]);
            }

            for (int i = 0; i < intensities.Length; i++)
            {
                intensities[i] /= max;
            }

            //calculate new intensities
            double mean = 0d;
            while (Math.Round(mean, 2) != 0.5)
            {
                mean = intensities.Sum() / intensities.Length;

                double theta = Math.Log(0.5) / Math.Log(mean);
                
                for (int i = 0; i < intensities.Length; i++)
                {
                    intensities[i] = Math.Pow(intensities[i], theta);
                }
            }

            //normalize to range
            double min = 255d;
            max = 0d;

            for (int i = 0; i < intensities.Length; i++)
            {
                min = Math.Min(min, intensities[i]);
                max = Math.Max(max, intensities[i]);
            }

            for (int i = 0; i < intensities.Length; i++)
            {
                intensities[i] = 255 * (intensities[i] - min) / (max - min);
            }

            //make new image
            double[][] result = { hsi[0], hsi[1], intensities };
            Bitmap res_img = result.HSI2RGB();
            return res_img;
        }
    }
}
